@@ -267,7 +267,8 @@ static int oss_dsp_hw_params(oss_dsp_t *dsp)
 			if (err < 0)
 				return err;
 			if (dsp->maxfrags > 0) {
-				unsigned int periods_max = dsp->maxfrags;
+				unsigned int periods_max = periods_min > dsp->maxfrags
+					? periods_min : dsp->maxfrags;
 				err = snd_pcm_hw_params_set_periods_max(pcm, hw,
 									&periods_max, 0);
 				if (err < 0)
@@ -470,6 +471,31 @@ int lib_oss_pcm_close(int fd)
 	return 0;
 }
 
+static int open_pcm(oss_dsp_t *dsp, const char *name, unsigned int pcm_mode,
+		    unsigned int streams)
+{
+	int k, result;
+
+	result = -ENODEV;
+	for (k = 0; k < 2; ++k) {
+		if (!(streams & (1 << k)))
+			continue;
+		result = snd_pcm_open(&dsp->streams[k].pcm, name, k, SND_PCM_NONBLOCK);
+		DEBUG("Opened PCM %s for stream %d (result = %d)\n", name, k, result);
+		if (result < 0) {
+			if (k == 1 && dsp->streams[0].pcm != NULL) {
+				dsp->streams[1].pcm = NULL;
+				streams &= ~(1 << SND_PCM_STREAM_CAPTURE);
+				result = 0;
+			}
+			break;
+		} else if (! pcm_mode)
+			/* reset the blocking mode */
+			snd_pcm_nonblock(dsp->streams[k].pcm, 0);
+	}
+	return result;
+}
+
 static int oss_dsp_open(int card, int device, int oflag, mode_t mode)
 {
 	oss_dsp_t *dsp;
@@ -551,39 +577,21 @@ static int oss_dsp_open(int card, int device, int oflag, mode_t mode)
 		if (result < 0)
 			goto _error;
 	}
-	for (k = 0; k < 2; ++k) {
-		if (!(streams & (1 << k)))
-			continue;
-		result = snd_pcm_open(&dsp->streams[k].pcm, name, k, pcm_mode);
-		if (result < 0) {
-			if (k == 1 && dsp->streams[0].pcm != NULL) {
-				dsp->streams[1].pcm = NULL;
-				streams &= ~(1 << SND_PCM_STREAM_CAPTURE);
-				result = 0;
-			}
-			break;
-		}
-	}
+	s = getenv("ALSA_OSS_PCM_DEVICE");
+	result = -ENODEV;
+	if (s && *s)
+		result = open_pcm(dsp, s, pcm_mode, streams);
+	if (result < 0)
+		result = open_pcm(dsp, name, pcm_mode, streams);
 	if (result < 0) {
-		result = 0;
-		for (k = 0; k < 2; ++k) {
-			if (dsp->streams[k].pcm) {
-				snd_pcm_close(dsp->streams[k].pcm);
-				dsp->streams[k].pcm = NULL;
-			}
-		}
 		/* try to open the default pcm as fallback */
 		if (card == 0 && (device == OSS_DEVICE_DSP || device == OSS_DEVICE_AUDIO))
 			strcpy(name, "default");
 		else
 			sprintf(name, "plughw:%d", card);
-		for (k = 0; k < 2; ++k) {
-			if (!(streams & (1 << k)))
-				continue;
-			result = snd_pcm_open(&dsp->streams[k].pcm, name, k, pcm_mode);
-			if (result < 0)
-				goto _error;
-		}
+		result = open_pcm(dsp, name, pcm_mode, streams);
+		if (result < 0)
+			goto _error;
 	}
 	result = oss_dsp_params(dsp);
 	if (result < 0)
@@ -1651,6 +1659,14 @@ int lib_oss_pcm_open(const char *file, int oflag, ...)
 			minor = (atoi(file + 9) << 4) | OSS_DEVICE_ADSP;
 		else if (!strncmp(file, "/dev/audio", 10))
 			minor = (atoi(file + 10) << 4) | OSS_DEVICE_AUDIO;
+		else if (!strncmp(file, "/dev/sound/dsp", 14))
+			minor = (atoi(file + 14) << 4) | OSS_DEVICE_DSP;
+		else if (!strncmp(file, "/dev/sound/dspW", 15))
+			minor = (atoi(file + 15) << 4) | OSS_DEVICE_DSPW;
+		else if (!strncmp(file, "/dev/sound/adsp", 15))
+			minor = (atoi(file + 15) << 4) | OSS_DEVICE_ADSP;
+		else if (!strncmp(file, "/dev/sound/audio", 16))
+			minor = (atoi(file + 16) << 4) | OSS_DEVICE_AUDIO;
 		else {
 			errno = ENOENT;
 			return -1;
